@@ -3,6 +3,8 @@
  * SPDX-License-Identifier: Apache-2.0
  */
 
+use crate::body::SdkBody;
+use crate::byte_stream::{error::Error, error::ErrorKind, ByteStream};
 use bytes::Bytes;
 use futures_core::ready;
 use http::HeaderMap;
@@ -15,10 +17,6 @@ use tokio::fs::File;
 use tokio::io::{self, AsyncReadExt, AsyncSeekExt};
 use tokio_util::io::ReaderStream;
 
-use crate::body::SdkBody;
-
-use super::{ByteStream, Error};
-
 // 4KB corresponds to the default buffer size used by Tokio's ReaderStream
 const DEFAULT_BUFFER_SIZE: usize = 4096;
 // By default, read files from their start
@@ -60,7 +58,7 @@ impl PathBody {
     }
 }
 
-/// Builder for creating [`ByteStreams`](crate::byte_stream::ByteStream) from a file/path, with full control over advanced options.
+/// Builder for creating [`ByteStreams`](ByteStream) from a file/path, with full control over advanced options.
 ///
 /// Example usage:
 /// ```no_run
@@ -86,8 +84,9 @@ impl PathBody {
 /// }
 /// # }
 /// ```
+#[allow(missing_debug_implementations)]
 pub struct FsBuilder {
-    file: Option<tokio::fs::File>,
+    file: Option<File>,
     path: Option<PathBuf>,
     length: Option<Length>,
     buffer_size: usize,
@@ -101,6 +100,7 @@ impl Default for FsBuilder {
 }
 
 /// The length (in bytes) to read. Determines whether or not a short read counts as an error.
+#[allow(missing_debug_implementations)]
 pub enum Length {
     /// Read this number of bytes exactly. Returns an error if the file is smaller than expected.
     Exact(u64),
@@ -137,7 +137,7 @@ impl FsBuilder {
     ///
     /// NOTE: The resulting ByteStream (after calling [build](FsBuilder::build)) will not be a retryable ByteStream.
     /// For a ByteStream that can be retried in the case of upstream failures, use [`FsBuilder::path`](FsBuilder::path).
-    pub fn file(mut self, file: tokio::fs::File) -> Self {
+    pub fn file(mut self, file: File) -> Self {
         self.file = Some(file);
         self
     }
@@ -169,7 +169,7 @@ impl FsBuilder {
         self
     }
 
-    /// Returns a [`ByteStream`](crate::byte_stream::ByteStream) from this builder.
+    /// Returns a [`ByteStream`](ByteStream) from this builder.
     pub async fn build(self) -> Result<ByteStream, Error> {
         if self.path.is_some() && self.file.is_some() {
             panic!("The 'file' and 'path' options on an FsBuilder are mutually exclusive but both were set. Please set only one")
@@ -181,19 +181,13 @@ impl FsBuilder {
         // notify users when file/chunk is smaller than expected.
         let file_length = self.get_file_size().await?;
         if offset > file_length {
-            return Err(Error(Box::new(io::Error::new(
-                io::ErrorKind::InvalidInput,
-                "offset must be less than or equal to file size but was greater than",
-            ))));
+            return Err(ErrorKind::OffsetLargerThanFileSize.into());
         }
 
         let length = match self.length {
             Some(Length::Exact(length)) => {
                 if length > file_length - offset {
-                    return Err(Error(Box::new(io::Error::new(
-                        io::ErrorKind::InvalidInput,
-                        "Length::Exact was larger than file size minus read offset",
-                    ))));
+                    return Err(ErrorKind::LengthLargerThanFileSizeMinusReadOffset.into());
                 }
                 length
             }
@@ -217,10 +211,7 @@ impl FsBuilder {
         } else if let Some(mut file) = self.file {
             // When starting from a `File`, we need to do our own seeking
             if offset != 0 {
-                let _s = file
-                    .seek(std::io::SeekFrom::Start(offset))
-                    .await
-                    .map_err(|err| Error(err.into()))?;
+                let _s = file.seek(io::SeekFrom::Start(offset)).await?;
             }
 
             let body = SdkBody::from_dyn(http_body::combinators::BoxBody::new(
@@ -234,20 +225,19 @@ impl FsBuilder {
     }
 
     async fn get_file_size(&self) -> Result<u64, Error> {
-        match self.path.as_ref() {
+        Ok(match self.path.as_ref() {
             Some(path) => tokio::fs::metadata(path).await,
             // If it's not path-based then it's file-based
             None => self.file.as_ref().unwrap().metadata().await,
         }
-        .map(|metadata| metadata.len())
-        .map_err(|err| Error(err.into()))
+        .map(|metadata| metadata.len())?)
     }
 }
 
 enum State {
     Unloaded(PathBuf),
     Loading(Pin<Box<dyn Future<Output = io::Result<File>> + Send + Sync + 'static>>),
-    Loaded(tokio_util::io::ReaderStream<io::Take<File>>),
+    Loaded(ReaderStream<io::Take<File>>),
 }
 
 impl Body for PathBody {
@@ -264,10 +254,10 @@ impl Body for PathBody {
                 State::Unloaded(ref path_buf) => {
                     let buf = path_buf.clone();
                     self.state = State::Loading(Box::pin(async move {
-                        let mut file = tokio::fs::File::open(&buf).await?;
+                        let mut file = File::open(&buf).await?;
 
                         if offset != 0 {
-                            let _s = file.seek(std::io::SeekFrom::Start(offset)).await?;
+                            let _s = file.seek(io::SeekFrom::Start(offset)).await?;
                         }
 
                         Ok(file)
@@ -358,8 +348,8 @@ mod test {
         assert_eq!(some_data.len(), 16384);
 
         assert_eq!(
-            ByteStream::new(body1).collect().await.unwrap().remaining(),
-            file_length as usize - some_data.len()
+            ByteStream::new(body1).collect().await.unwrap().remaining() as u64,
+            file_length - some_data.len() as u64
         );
     }
 
@@ -577,7 +567,7 @@ mod test {
 
         let file_size = file.as_file().metadata().unwrap().len();
         // Check that our in-memory copy has the same size as the file
-        assert_eq!(file_size as usize, in_memory_copy_of_file_contents.len());
+        assert_eq!(file_size, in_memory_copy_of_file_contents.len() as u64);
         let file_path = file.path().to_path_buf();
         let chunks = 7;
         let chunk_size = file_size / chunks;
@@ -586,7 +576,7 @@ mod test {
         for i in 0..chunks {
             let length = if i == chunks - 1 {
                 // If we're on the last chunk, the length to read might be less than a whole chunk.
-                // We substract the size of all previous chunks from the total file size to get the
+                // We subtract the size of all previous chunks from the total file size to get the
                 // size of the final chunk.
                 file_size - (i * chunk_size)
             } else {
