@@ -7,13 +7,14 @@
 
 use std::error::Error as StdError;
 
-mod input;
-mod output;
+mod receiver;
+mod sender;
 
+/// A generic, boxed error that's `Send`, `Sync`, and `'static`.
 pub type BoxError = Box<dyn StdError + Send + Sync + 'static>;
 
 #[doc(inline)]
-pub use input::{EventStreamInput, MessageStreamAdapter};
+pub use sender::{EventStreamSender, MessageStreamAdapter, MessageStreamError};
 
 #[doc(inline)]
-pub use output::{Error, RawMessage, Receiver};
+pub use receiver::{RawMessage, Receiver, ReceiverError};
