@@ -5,66 +5,68 @@
 
 //! Utilities for parsing information from headers
 
+use aws_smithy_types::date_time::Format;
+use aws_smithy_types::primitive::Parse;
+use aws_smithy_types::DateTime;
+use http::header::{HeaderMap, HeaderName, HeaderValue, ValueIter};
 use std::borrow::Cow;
 use std::convert::TryFrom;
 use std::error::Error;
 use std::fmt;
-use std::fmt::{Display, Formatter};
 use std::str::FromStr;
 
-use http::header::{HeaderMap, HeaderName, HeaderValue, ValueIter};
-
-use aws_smithy_types::date_time::Format;
-use aws_smithy_types::primitive::Parse;
-use aws_smithy_types::DateTime;
-
-#[derive(Debug, Eq, PartialEq)]
-#[non_exhaustive]
+/// An error was encountered while parsing a header
+#[derive(Debug)]
 pub struct ParseError {
-    message: Option<Cow<'static, str>>,
+    message: Cow<'static, str>,
+    source: Option<Box<dyn Error + Send + Sync + 'static>>,
 }
 
 impl ParseError {
-    #[allow(clippy::new_without_default)]
-    pub fn new() -> Self {
-        Self { message: None }
-    }
-
-    pub fn new_with_message(message: impl Into<Cow<'static, str>>) -> Self {
+    /// Create a new parse error with the given `message`
+    pub fn new(message: impl Into<Cow<'static, str>>) -> Self {
         Self {
-            message: Some(message.into()),
+            message: message.into(),
+            source: None,
+        }
+    }
+
+    fn with_source(self, source: impl Into<Box<dyn Error + Send + Sync + 'static>>) -> Self {
+        Self {
+            source: Some(source.into()),
+            ..self
         }
     }
 }
 
-impl Display for ParseError {
-    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
-        write!(f, "Output failed to parse in headers")?;
-        if let Some(message) = &self.message {
-            write!(f, ". {}", message)?;
-        }
-        Ok(())
+impl fmt::Display for ParseError {
+    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
+        write!(f, "output failed to parse in headers: {}", self.message)
     }
 }
 
-impl Error for ParseError {}
+impl Error for ParseError {
+    fn source(&self) -> Option<&(dyn Error + 'static)> {
+        self.source.as_ref().map(|err| err.as_ref() as _)
+    }
+}
 
 /// Read all the dates from the header map at `key` according the `format`
 ///
 /// This is separate from `read_many` below because we need to invoke `DateTime::read` to take advantage
 /// of comma-aware parsing
 pub fn many_dates(
-    values: ValueIter<HeaderValue>,
+    values: ValueIter<'_, HeaderValue>,
     format: Format,
 ) -> Result<Vec<DateTime>, ParseError> {
     let mut out = vec![];
     for header in values {
         let mut header = header
             .to_str()
-            .map_err(|_| ParseError::new_with_message("header was not valid utf-8 string"))?;
+            .map_err(|_| ParseError::new("header was not valid utf-8 string"))?;
         while !header.is_empty() {
             let (v, next) = DateTime::read(header, format, ',').map_err(|err| {
-                ParseError::new_with_message(format!("header could not be parsed as date: {}", err))
+                ParseError::new(format!("header could not be parsed as date: {}", err))
             })?;
             out.push(v);
             header = next;
@@ -73,6 +75,8 @@ pub fn many_dates(
     Ok(out)
 }
 
+/// Returns an iterator over pairs where the first element is the unprefixed header name that
+/// starts with the input `key` prefix, and the second element is the full header name.
 pub fn headers_for_prefix<'a>(
     headers: &'a http::HeaderMap,
     key: &'a str,
@@ -84,29 +88,33 @@ pub fn headers_for_prefix<'a>(
         .map(move |h| (&h.as_str()[key.len()..], h))
 }
 
+/// Convert a `HeaderValue` into a `Vec<T>` where `T: FromStr`
 pub fn read_many_from_str<T: FromStr>(
-    values: ValueIter<HeaderValue>,
-) -> Result<Vec<T>, ParseError> {
+    values: ValueIter<'_, HeaderValue>,
+) -> Result<Vec<T>, ParseError>
+where
+    T::Err: Error + Send + Sync + 'static,
+{
     read_many(values, |v: &str| {
-        v.parse()
-            .map_err(|_err| ParseError::new_with_message("failed during FromString conversion"))
+        v.parse().map_err(|err| {
+            ParseError::new("failed during `FromString` conversion").with_source(err)
+        })
     })
 }
 
-pub fn read_many_primitive<T: Parse>(values: ValueIter<HeaderValue>) -> Result<Vec<T>, ParseError> {
+/// Convert a `HeaderValue` into a `Vec<T>` where `T: Parse`
+pub fn read_many_primitive<T: Parse>(
+    values: ValueIter<'_, HeaderValue>,
+) -> Result<Vec<T>, ParseError> {
     read_many(values, |v: &str| {
-        T::parse_smithy_primitive(v).map_err(|primitive| {
-            ParseError::new_with_message(format!(
-                "failed reading a list of primitives: {}",
-                primitive
-            ))
-        })
+        T::parse_smithy_primitive(v)
+            .map_err(|err| ParseError::new("failed reading a list of primitives").with_source(err))
     })
 }
 
 /// Read many comma / header delimited values from HTTP headers for `FromStr` types
 fn read_many<T>(
-    values: ValueIter<HeaderValue>,
+    values: ValueIter<'_, HeaderValue>,
     f: impl Fn(&str) -> Result<T, ParseError>,
 ) -> Result<Vec<T>, ParseError> {
     let mut out = vec![];
@@ -125,24 +133,28 @@ fn read_many<T>(
 ///
 /// This function does not perform comma splitting like `read_many`
 pub fn one_or_none<T: FromStr>(
-    mut values: ValueIter<HeaderValue>,
-) -> Result<Option<T>, ParseError> {
+    mut values: ValueIter<'_, HeaderValue>,
+) -> Result<Option<T>, ParseError>
+where
+    T::Err: Error + Send + Sync + 'static,
+{
     let first = match values.next() {
         Some(v) => v,
         None => return Ok(None),
     };
-    let value = std::str::from_utf8(first.as_bytes())
-        .map_err(|_| ParseError::new_with_message("invalid utf-8"))?;
+    let value =
+        std::str::from_utf8(first.as_bytes()).map_err(|_| ParseError::new("invalid utf-8"))?;
     match values.next() {
         None => T::from_str(value.trim())
-            .map_err(|_| ParseError::new())
+            .map_err(|err| ParseError::new("failed to parse string").with_source(err))
             .map(Some),
-        Some(_) => Err(ParseError::new_with_message(
+        Some(_) => Err(ParseError::new(
             "expected a single value but found multiple",
         )),
     }
 }
 
+/// Given an HTTP request, set a request header if that header was not already set.
 pub fn set_request_header_if_absent<V>(
     request: http::request::Builder,
     key: HeaderName,
@@ -163,6 +175,7 @@ where
     }
 }
 
+/// Given an HTTP response, set a response header if that header was not already set.
 pub fn set_response_header_if_absent<V>(
     response: http::response::Builder,
     key: HeaderName,
@@ -229,7 +242,7 @@ mod parse_multi_header {
         let next_delim = input.iter().position(|&b| b == b',').unwrap_or(input.len());
         let (first, next) = input.split_at(next_delim);
         let first = std::str::from_utf8(first)
-            .map_err(|_| ParseError::new_with_message("header was not valid utf8"))?;
+            .map_err(|_| ParseError::new("header was not valid utf-8"))?;
         Ok((Cow::Borrowed(first), then_comma(next).unwrap()))
     }
 
@@ -239,10 +252,10 @@ mod parse_multi_header {
         for index in 0..input.len() {
             match input[index] {
                 b'"' if index == 0 || input[index - 1] != b'\\' => {
-                    let mut inner =
-                        Cow::Borrowed(std::str::from_utf8(&input[0..index]).map_err(|_| {
-                            ParseError::new_with_message("header was not valid utf8")
-                        })?);
+                    let mut inner = Cow::Borrowed(
+                        std::str::from_utf8(&input[0..index])
+                            .map_err(|_| ParseError::new("header was not valid utf-8"))?,
+                    );
                     inner = replace(inner, "\\\"", "\"");
                     inner = replace(inner, "\\\\", "\\");
                     let rest = then_comma(&input[(index + 1)..])?;
@@ -251,7 +264,7 @@ mod parse_multi_header {
                 _ => {}
             }
         }
-        Err(ParseError::new_with_message(
+        Err(ParseError::new(
             "header value had quoted value without end quote",
         ))
     }
@@ -262,7 +275,7 @@ mod parse_multi_header {
         } else if s.starts_with(b",") {
             Ok(&s[1..])
         } else {
-            Err(ParseError::new_with_message("expected delimiter `,`"))
+            Err(ParseError::new("expected delimiter `,`"))
         }
     }
 }
@@ -296,7 +309,7 @@ pub fn quote_header_value<'a>(value: impl Into<Cow<'a, str>>) -> Cow<'a, str> {
 
 /// Given two [`HeaderMap`][HeaderMap]s, merge them together and return the merged `HeaderMap`. If the
 /// two `HeaderMap`s share any keys, values from the right `HeaderMap` be appended to the left `HeaderMap`.
-pub(crate) fn append_merge_header_maps(
+pub fn append_merge_header_maps(
     mut lhs: HeaderMap<HeaderValue>,
     rhs: HeaderMap<HeaderValue>,
 ) -> HeaderMap<HeaderValue> {
@@ -335,6 +348,7 @@ mod test {
     };
 
     use super::quote_header_value;
+    use aws_smithy_types::error::display::DisplayErrorContext;
 
     #[test]
     fn put_on_request_if_absent() {
@@ -386,13 +400,18 @@ mod test {
                 .expect("valid"),
             vec![0.0, f32::INFINITY, f32::NEG_INFINITY, 5555.5]
         );
-        assert_eq!(
-            read_many_primitive::<f32>(test_request.headers().get_all("X-Float-Error").iter())
-                .expect_err("invalid"),
-            ParseError::new_with_message(
-                "failed reading a list of primitives: failed to parse input as f32"
+        let message = format!(
+            "{}",
+            DisplayErrorContext(
+                read_many_primitive::<f32>(test_request.headers().get_all("X-Float-Error").iter())
+                    .expect_err("invalid")
             )
-        )
+        );
+        let expected = "output failed to parse in headers: failed reading a list of primitives: failed to parse input as f32";
+        assert!(
+            message.starts_with(expected),
+            "expected '{message}' to start with '{expected}'"
+        );
     }
 
     #[test]
@@ -602,10 +621,8 @@ mod test {
 
         let merged_header_map =
             append_merge_header_maps(left_hand_side_headers, right_hand_side_headers);
-        let actual_merged_values: Vec<_> = merged_header_map
-            .get_all(header_name.clone())
-            .into_iter()
-            .collect();
+        let actual_merged_values: Vec<_> =
+            merged_header_map.get_all(header_name).into_iter().collect();
 
         let expected_merged_values = vec![left_header_value, right_header_value];
 
@@ -628,10 +645,8 @@ mod test {
 
         let merged_header_map =
             append_merge_header_maps(left_hand_side_headers, right_hand_side_headers);
-        let actual_merged_values: Vec<_> = merged_header_map
-            .get_all(header_name.clone())
-            .into_iter()
-            .collect();
+        let actual_merged_values: Vec<_> =
+            merged_header_map.get_all(header_name).into_iter().collect();
 
         let expected_merged_values =
             vec![left_header_value_1, left_header_value_2, right_header_value];
@@ -653,10 +668,8 @@ mod test {
 
         let merged_header_map =
             append_merge_header_maps(left_hand_side_headers, right_hand_side_headers);
-        let actual_merged_values: Vec<_> = merged_header_map
-            .get_all(header_name.clone())
-            .into_iter()
-            .collect();
+        let actual_merged_values: Vec<_> =
+            merged_header_map.get_all(header_name).into_iter().collect();
 
         let expected_merged_values = vec![right_header_value_1, right_header_value_2];
 
